package ru.yandex.intranet.d.util.dispatchers

import com.google.common.util.concurrent.ThreadFactoryBuilder
import kotlinx.coroutines.CoroutineExceptionHandler
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.ExecutorCoroutineDispatcher
import kotlinx.coroutines.Job
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.asCoroutineDispatcher
import kotlinx.coroutines.cancel
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.slf4j.MDCContext
import mu.KotlinLogging
import org.springframework.context.annotation.Profile
import org.springframework.stereotype.Component
import java.util.concurrent.ScheduledExecutorService
import java.util.concurrent.ScheduledThreadPoolExecutor
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicReference
import javax.annotation.PreDestroy

private val logger = KotlinLogging.logger {}

/**
 * Background heavy dispatcher blocking stub.
 *
 * @author Dmitriy Timashov <dm-tim@yandex-team.ru>
 */
@Component("backgroundHeavyDispatcher")
@Profile("integration-tests", "test-recipe")
class BackgroundHeavyDispatcherBlockingStub: CustomDispatcher {

    private val scheduler: ScheduledExecutorService
    private val dispatcher: ExecutorCoroutineDispatcher
    private val coroutinesScope: CoroutineScope

    init {
        val threadFactory = ThreadFactoryBuilder()
            .setDaemon(true)
            .setNameFormat("background-heavy-pool-%d")
            .setUncaughtExceptionHandler { t: Thread, e: Throwable? ->
                logger.error(e) { "Uncaught exception in background heavy pool thread $t" }
            }
            .build()
        val scheduledThreadPoolExecutor = ScheduledThreadPoolExecutor(
            (Runtime.getRuntime().availableProcessors() / 4).coerceAtLeast(2), threadFactory)
        scheduledThreadPoolExecutor.removeOnCancelPolicy = true
        scheduler = scheduledThreadPoolExecutor
        dispatcher = scheduledThreadPoolExecutor.asCoroutineDispatcher()
        val exceptionHandler = CoroutineExceptionHandler { _, exception ->
            logger.error(exception) { "Uncaught exception in background heavy pool" }
        }
        coroutinesScope = CoroutineScope(dispatcher + exceptionHandler + SupervisorJob())
    }

    @PreDestroy
    fun preDestroy() {
        logger.info("Stopping background heavy dispatcher...")
        coroutinesScope.cancel("Application context shutdown")
        dispatcher.close()
        try {
            scheduler.awaitTermination(1, TimeUnit.SECONDS)
        } catch (e: InterruptedException) {
            Thread.currentThread().interrupt()
        }
        scheduler.shutdownNow()
        logger.info("Stopped background heavy dispatcher")
    }

    override fun launch(block: suspend () -> Unit): Job {
        val job = coroutinesScope.launch(MDCContext()) { block() }
        val exception = AtomicReference<Throwable>(null)
        job.invokeOnCompletion { e -> exception.set(e) }
        runBlocking {
            job.join()
        }
        if (exception.get() != null) {
            throw exception.get()
        }
        return job
    }

}
